# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# (C) British Crown Copyright 2017-2020 Met Office.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

import numpy as np
from scipy.stats.distributions import t
import matplotlib.pyplot as plt
import iris
import iris.analysis as ia
import cartopy.crs as ccrs
import doctest
import os.path
import catnip.config as conf


def linear_regress(xi, yi):
    """
    Solves y = mx + c by returning the
    least squares solution to a linear matrix
    equation. Expects two numpy arrays of dimension 1.

    args
    ----
    xi: numpy array of dimension 1
    yi: dependant variable, numpy array of dimension 1

    Returns
    -------
    grad: gradient i.e. m in y = mx + c
    intcp: intercept i.e. c in y = mx + c
    xpts: the min and max value of xi, used to plot line of best fit
    ypts: the min and max of the y solutions to line of best fit for plotting

    Notes
    ------
    TODO: use sklearn to perform regression


    A simple example:

    >>> x = np.array([1, 4, 2, 7, 0, 6, 3, 3, 1, 9])
    >>> y = np.array([5, 6, 2, 9, 1, 4, 7, 8, 2, 6])
    >>> grad, intcp, xp, yp, sum_res = linear_regress(x, y)
    >>> print('gradient', "{:.2f}".format(grad))
    gradient 0.54
    >>> print('intercept', "{:.2f}".format(intcp))
    intercept 3.07
    """

    if np.shape(xi) != np.shape(yi):
        raise ValueError(
            "The input fields do not have the same shape, \
            {} and {}".format(
                str(np.shape(xi)), str(np.shape(yi))
            )
        )

    elif np.ndim(xi) != 1 or np.ndim(yi) != 1:
        raise ValueError(
            "xi and yi must have dinemsion 1, not xi \
            {} and yi {}".format(
                str(np.ndim(xi)), str(np.ndim(yi))
            )
        )
    else:
        # Return the least-squares solution to a linear matrix equation
        regression = np.linalg.lstsq(
            np.vstack([xi, np.ones(len(xi))]).T, yi, rcond=None
        )

        # Get the slope and the intercept
        grad, intcp = regression[0]

        # sum of residuals: squared Euclidean 2-norm for each column in yi - xi*x
        sum_res = regression[1]

        xmin, xmax = np.min(xi), np.max(xi)
        xpts = [xmin, xmax]
        ypts = grad * np.array(xpts) + intcp

    return grad, intcp, xpts, ypts, sum_res


def ci_interval(xi, yi, alpha=0.05):

    """
    Calculates  Confidence interval (default 95%) parameters for two
    numpy arrays of dimension 1. Returns include the gradient
    and intercept of the confidence interval.

    It also returns vectors for confidence interval plotting:
    -  the high/low confidence interval of the slope
    -  the high low curves of confidence interval of yi.

    args
    ----
    xi: numpy array of dimension 1
    yi: dependant variable, numpy array of dimension 1
    alpha: required confidence interval (e.g. 0.05 for 95%). Default is 0.05.

    Returns
    -------
    slope_conf_int: Gradient of confidence interval
    intcp_conf_int: Intercept (calculated using mean(yi)) of confidence interval
    xpts: the min and max value of xi.
    slope_lo_pts: for plotting, min and max y value of lower bound of CI of slope
    slope_hi_pts: for plotting, min and max y value of upper bound of CI of slope
    xreg: x values spanning xmin to xmax linearly spaced, for plotting against  yi CI curve.
    y_conf_int_lo: for plotting, lower bound of CI region for yi
    y_conf_int_hi: for plotting, upper bound of CI region for yi

    Notes
    -----
    Parameters have been calculated using von Storch & Zwiers Statisical Analysis in Climate Research
    Sect.8.3.7 and 8.3.10

    A simple example:

    >>> x = np.array([1, 4, 2, 7, 0, 6, 3, 3, 1, 9])
    >>> y = np.array([5, 6, 2, 9, 1, 4, 7, 8, 2, 6])
    >>> slope_conf_int, intcp_conf_int, xpts, slope_lo_pts, slope_hi_pts, \
            xreg, y_conf_int_lo, y_conf_int_hi = ci_interval(x, y)
    Calculating the 95.0% confidence interval
    >>> print('CI gradient', "{:.2f}".format(slope_conf_int))
    CI gradient 0.62
    >>> print('CI intercept', "{:.2f}".format(intcp_conf_int))
    CI intercept 2.81

    """

    if xi.shape != yi.shape:
        raise ValueError(
            "The input fields do not have the same shape, \
            {} and {}".format(
                str(xi.shape), str(yi.shape)
            )
        )

    print(("Calculating the {}% confidence interval".format(str((1 - alpha) * 100))))
    # number of points
    n = len(xi)
    # degrees of freedom
    dof = n - 2  # assuming 2 parameters
    # student-t value for the dof and confidence level
    # Note, ppf - Percent point function (inverse of cdf - percentiles).
    t_val = t.ppf(1.0 - alpha / 2.0, dof)

    # Return the least-squares solution to a linear matrix equation
    slope, intcp, xp, yp, sum_res = linear_regress(xi, yi)

    # If yi is 1-dimensional, sum_res is a (1,) shape array.
    # Use this to calculate confidence intervals for plotting.
    if sum_res.shape == (1,):
        # standard error
        sd_err = np.sqrt(sum_res / (dof))[0]
        # calculate the sum of the squares of the difference
        # between each x and the mean x value
        sxx = np.sum((xi - np.mean(xi)) ** 2)
        if sxx == 0.0:
            raise ValueError("Sum of squares of difference is 0")

        # CI of slope: Formulated from vonStorch & Zwiers Sect.8.3.7
        slope_conf_int = (t_val * sd_err) / np.sqrt(sxx)
        # lower bound of slope using CI interval
        slope_lo = slope - slope_conf_int
        # upper bound of slope using CI interval
        slope_hi = slope + slope_conf_int
        xmin, xmax = np.min(xi), np.max(xi)
        # for plotting CI slope you want
        xpts = xmin, xmax
        slope_lo_pts = np.mean(yi) + slope_lo * ([xmin, xmax] - np.mean(xi))
        slope_hi_pts = np.mean(yi) + slope_hi * ([xmin, xmax] - np.mean(xi))

        # get regularly spaced np array of x points
        xreg = np.linspace(xmin, xmax, 101)
        # Population yi CI: Formulated from vonStorch & Zwiers Sect.8.3.10
        # CI for the mean of the response variable
        yfact = np.sqrt((1.0 / n) + (((xreg - np.mean(xi)) ** 2) / sxx))
        # 95% CI
        ymean_conf_int = (t_val * sd_err) * yfact
        # get the lines for plotting ymean CI
        y_conf_int_hi = (slope * xreg) + intcp + ymean_conf_int
        y_conf_int_lo = (slope * xreg) + intcp - ymean_conf_int

        # Intercept CI, using population y CI at x=0
        # Note, this isn't meaningful if x=0 is not physically meaningful
        yfact0 = np.sqrt((1.0 / n) + ((np.mean(xi) ** 2) / sxx))
        intcp_conf_int = (t_val * sd_err) * yfact0
        intcp_lo = intcp - intcp_conf_int
        intcp_hi = intcp + intcp_conf_int

        return (
            slope_conf_int,
            intcp_conf_int,
            xpts,
            slope_lo_pts,
            slope_hi_pts,
            xreg,
            y_conf_int_lo,
            y_conf_int_hi,
        )


def regrid_to_target(cube, target_cube, method="linear", extrap="mask", mdtol=0.5):
    """
    Takes in two cubes, and regrids one onto the grid
    of the other. Optional arguments include the method
    of regridding, default is linear. The method of
    extrapolation (if it is required), the default is
    to mask the data, even if the source data is not
    a masked array. And, if the method is areaweighted,
    choose a missing data tolerance, default is 0.5.
    For full info, see https://scitools.org.uk/iris/docs/latest/userguide/interpolation_and_regridding.html

    args
    ----
    cube: cube you want to regrid
    target_cube: cube on the target grid
    method: method of regridding, options are 'linear', 'nearest' and 'areaweighted'.
    extrap: extraopolation mode, options are 'mask', 'nan', 'error' and 'nanmask'
    mdtol: tolerated fraction of masked data in any given target grid-box, only used if method='areaweighted', between 0 and 1.

    Returns
    -------
    cube_reg: input cube on the grid of target_cube

    Notes
    -----
    areaweighted is VERY picky, it will not allow you to regrid using this method if the two input cubes are not on the
    same coordinate system, and both input grids must also contain monotonic, bounded, 1D spatial coordinates.

    An example:

    >>> file1 = os.path.join(conf.DATA_DIR, 'gcm_monthly.pp')
    >>> file2 = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube = iris.load_cube(file1, 'air_temperature')
    >>> tgrid = iris.load_cube(file2, 'air_temperature')
    >>> cube_reg = regrid_to_target(cube, tgrid)
    regridding from GeogCS(6371229.0) to RotatedGeogCS(39.25, 198.0, ellipsoid=GeogCS(6371229.0)) using method linear
    >>> print(cube.shape, tgrid.shape)
    (145, 192) (2, 433, 444)
    >>> print(cube_reg.shape)
    (433, 444)
    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    if not isinstance(target_cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    target_cs = target_cube.coord(axis="x").coord_system
    orig_cs = cube.coord(axis="x").coord_system

    # get coord names for cube
    # Longitude
    xcoord = cube.coord(axis="X", dim_coords=True)
    # Latitude
    ycoord = cube.coord(axis="Y", dim_coords=True)
    # get coord names for target_cube
    # Longitude
    t_xcoord = target_cube.coord(axis="X", dim_coords=True)
    # Latitude
    t_ycoord = target_cube.coord(axis="Y", dim_coords=True)

    if method == "linear":
        scheme = iris.analysis.Linear(extrapolation_mode=extrap)

    if method == "nearest":
        scheme = iris.analysis.Nearest(extrapolation_mode=extrap)

    # areaweighted is VERY picky and often can't be used.
    if method == "areaweighted":
        if not cube.coord(xcoord).has_bounds():
            print("Input cube to be regrided does not have lon bounds, guessing . . . ")
            cube.coord(xcoord).guess_bounds()
        if not cube.coord(ycoord).has_bounds():
            print("Input cube to be regrided does not have lat bounds, guessing . . . ")
            cube.coord(ycoord).guess_bounds()

        if not target_cube.coord(t_xcoord).has_bounds():
            print("Input target_cube does not have lon bounds, guessing . . . ")
            target_cube.coord(t_xcoord).guess_bounds()
        if not target_cube.coord(t_ycoord).has_bounds():
            print("Input target_cube does not have lat bounds, guessing . . . ")
            target_cube.coord(t_ycoord).guess_bounds()

        if target_cs != orig_cs:
            raise ValueError("The input cubes must have the same coordinate system")
        scheme = iris.analysis.AreaWeighted(mdtol=mdtol)

    print(
        "regridding from {} to {} using method {}".format(
            str(orig_cs), str(target_cs), method
        )
    )

    cube_reg = cube.regrid(target_cube, scheme)

    return cube_reg


def set_regridder(cube, target_cube, method="linear", extrap="mask", mdtol=0.5):
    """
    Takes in two cubes, and sets up a regridder, mapping
    one cube to another. The most computationally expensive
    part of regridding is setting up the re-gridder, if you
    have a lot of cubes to re-grid, this will save time.
    Optional arguments include the method of regridding,
    default is linear.The method of extrapolation (if
    it is required), the default is to mask the data,
    even if the source data is not a masked array. And,
    if the method is areaweighted, choose a missing data
    tolerance, default is 0.5.


    Note: areaweighted is VERY picky, it will not allow you to regrid using
    this method if the two input cubes are not on the same coordinate system,
    and both input grids must also contain monotonic, bounded, 1D spatial coordinates.

    args
    ----
    cube: cube you want to regrid
    target_cube: cube on the target grid
    method: method of regridding, options are 'linear', 'nearest' and 'areaweighted'.
    extrap: extraopolation mode, options are 'mask', 'nan', 'error' and 'nanmask'
    mdtol: tolerated fraction of masked data in any given target grid-box, only used if method='areaweighted', between 0 and 1.

    Returns
    -------
    regridder: a cached regridder which can be used on any iris cube which has the same grid as cube.


    Notes
    -----
    See https://scitools.org.uk/iris/docs/latest/userguide/interpolation_and_regridding.html for more information

    An example:

    >>> file1 = os.path.join(conf.DATA_DIR, 'gcm_monthly.pp')
    >>> file2 = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube = iris.load_cube(file1, 'air_temperature')
    >>> tgrid = iris.load_cube(file2, 'air_temperature')
    >>> regridder = set_regridder(cube, tgrid)
    >>> cube2 = iris.load_cube(file1, 'cloud_area_fraction')
    >>> print(cube2.shape)
    (145, 192)
    >>> regridder(cube2)
    <iris 'Cube' of cloud_area_fraction / (1) (grid_latitude: 433; grid_longitude: 444)>
    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    if not isinstance(target_cube, iris.cube.Cube):
        raise TypeError("Target_cube is not of type cube")

    target_cs = target_cube.coord(axis="x").coord_system
    orig_cs = cube.coord(axis="x").coord_system

    # get coord names for cube
    # Longitude
    xcoord = cube.coord(axis="X", dim_coords=True)
    # Latitude
    ycoord = cube.coord(axis="Y", dim_coords=True)
    # get coord names for target_cube
    # Longitude
    t_xcoord = target_cube.coord(axis="X", dim_coords=True)
    # Latitude
    t_ycoord = target_cube.coord(axis="Y", dim_coords=True)

    if method == "linear":
        regridder = iris.analysis.Linear(extrapolation_mode=extrap).regridder(
            cube, target_cube
        )

    if method == "nearest":
        regridder = iris.analysis.Nearest(extrapolation_mode=extrap).regridder(
            cube, target_cube
        )

    if method == "areaweighted":
        if not cube.coord(xcoord).has_bounds():
            print("Input cube to be regrided does not have lon bounds, guessing . . . ")
            cube.coord(xcoord).guess_bounds()
        if not cube.coord(ycoord).has_bounds():
            print("Input cube to be regrided does not have lat bounds, guessing . . . ")
            cube.coord(ycoord).guess_bounds()

        if not target_cube.coord(t_xcoord).has_bounds():
            print("Input target_cube does not have lon bounds, guessing . . . ")
            target_cube.coord(t_xcoord).guess_bounds()
        if not target_cube.coord(t_ycoord).has_bounds():
            print("Input target_cube does not have lat bounds, guessing . . . ")
            target_cube.coord(t_ycoord).guess_bounds()

        if target_cs != orig_cs:
            raise ValueError("The input cubes must have the same coordinate system")

        regridder = iris.analysis.AreaWeighted(mdtol=mdtol).regridder(cube, target_cube)

    return regridder


def seas_time_stat(
    cube,
    seas_mons=[[3, 4, 5], [6, 7, 8], [9, 10, 11], [12, 1, 2]],
    metric="mean",
    pc=[],
    years=[],
    ext_area=[],
):
    """
    Takes in a cube and calculates a seasonal metric. Defaults to
    mean of 'mam','jja','son' and 'djf' over the whole time span
    of the cube. However, the user can specify seasons by month
    number i.e. ond=[10,11,12], the start and end years, the metric
    calculated and a lat lon region to extract for the calculation.

    args
    ----
    cube: Input cube that must contain a coordinate called 'time'
    seas_mons: list of seasons to calculate the metric over,
             defaults to seas_mons=[[3,4,5],[6,7,8],[9,10,11],[12,1,2]].
    metric: string, optional argument, defaults to 'mean', but can
          be 'mean', 'std_dev', 'min', 'max' or 'percentile'
    pc: optional argument, percentile level to calculate,
       must be an integer and must be set if metric='percentile'.
    years: list of start and end year, default is
           the whole time span of the cube.
    ext_area: optional argument, if set expects a list of
            the form [lonmin, lonmax, latmin, latmax].

    Returns
    -------
    cube_list: a cube list containing one cube per
             season of the calculated metric

    Notes
    -----
    The main draw back is that the function does not take
    account for seasons that span more than one year
    i.e. djf, and will calculate the metric over all months
    that meet the season criteria. For calculation
    where continuous seasons are important
    iris.aggregated_by is better.

    See an example:

    >>> file1 = os.path.join(conf.DATA_DIR, 'mslp.daily.rcm.viet.nc')
    >>> file2 = os.path.join(conf.DATA_DIR, 'FGOALS-g2_ua@925_nov.nc')
    >>> # load a rcm cube
    ... cube = iris.load_cube(file1)
    >>> seas_min_cubelist = seas_time_stat(cube, metric='min', years=[2000,2000])
    Calculating min for 2000-2000 mam
    Calculating min for 2000-2000 jja
    Calculating min for 2000-2000 son
    Calculating min for 2000-2000 djf
    >>> seas_mean_cubelist = seas_time_stat(cube, seas_mons=[[1,2],[6,7,8],[6,7,8,9],[10,11]], ext_area=[340, 350, 0,10])
    WARNING - the cube is on a rotated pole, the area you extract might not be where you think it is! You can use regular_point_to_rotated to check your ext_area lat and lon
    Calculating mean for 2000-2001 jf
    Calculating mean for 2000-2001 jja
    Calculating mean for 2000-2001 jjas
    Calculating mean for 2000-2001 on
    >>> # now load a gcm cube
    ... cube2 = iris.load_cube(file2)
    >>> seas_pc_cubelist = seas_time_stat(cube2, seas_mons=[[11]], metric='percentile', pc=95, ext_area=[340, 350, 0,10])
    Calculating percentile for 490-747 n
    >>> # print an example of the output
    ... print(seas_mean_cubelist[1])
    air_pressure_at_sea_level / (Pa)    (grid_latitude: 46; grid_longitude: 23)
         Dimension coordinates:
              grid_latitude                           x                   -
              grid_longitude                          -                   x
         Scalar coordinates:
              season: jja
              season_fullname: junjulaug
              time: 2000-07-16 00:00:00, bound=(2000-06-01 00:00:00, 2000-09-01 00:00:00)
         Attributes:
              Conventions: CF-1.5
              STASH: m01s16i222
              source: Data from Met Office Unified Model
         Cell methods:
              mean: time (1 hour)
              mean: time
    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # check the cube contains a coordinate called time
    coord_names = [coord.name() for coord in cube.coords(dim_coords=True)]

    if "time" not in coord_names:
        raise iris.exceptions.CoordinateNotFoundError(
            "No coordinate called 'time' in cube"
        )

    else:
        # if the start and end years are not defined by the user
        # default to using the whole time span of the cube
        if not years:
            time_info = cube.coord("time")
            if not cube.coord("time").has_bounds():
                raise Exception(
                    "Coordinate 'time' does not have bounds. Add bounds using the add_bounds function."
                )
            years = [
                time_info.units.num2date(time_info.bounds[0][0]).year,
                time_info.units.num2date(time_info.bounds[-1][1]).year,
            ]

        if ext_area:
            # check the coordinate system of the cube
            cs_str = str(cube.coord_system())
            if cs_str.find("Rotated") != -1:
                print(
                    "WARNING - the cube is on a rotated pole, the area you extract might not be where you think it is! You can use regular_point_to_rotated to check your ext_area lat and lon"
                )
            if len(ext_area) != 4:
                raise IndexError(
                    "area to extract must contain 4 values, currently contains {}".format(
                        str(len(ext_area))
                    )
                )
            else:
                if "grid_latitude" in coord_names:
                    cube = cube.intersection(
                        grid_longitude=(ext_area[0], ext_area[1]),
                        grid_latitude=(ext_area[2], ext_area[3]),
                    )
                elif "latitude" in coord_names:
                    cube = cube.intersection(
                        longitude=(ext_area[0], ext_area[1]),
                        latitude=(ext_area[2], ext_area[3]),
                    )
                else:
                    raise IndexError(
                        "Neither latitude nor grid_latitude coordinates in cube, can't extract area"
                    )

        # dictionary of month number to month letter, used to make strings
        # of season names e.g. 'jja'
        month_dict = {
            1: "j",
            2: "f",
            3: "m",
            4: "a",
            5: "m",
            6: "j",
            7: "j",
            8: "a",
            9: "s",
            10: "o",
            11: "n",
            12: "d",
        }
        month_fullname_dict = {
            1: "jan",
            2: "feb",
            3: "mar",
            4: "apr",
            5: "may",
            6: "jun",
            7: "jul",
            8: "aug",
            9: "sep",
            10: "oct",
            11: "nov",
            12: "dec",
        }

        cube_list = iris.cube.CubeList()

        # loop round the seasons
        for season in seas_mons:
            # make strings of the season
            seas_str = []
            seas_fullname_str = []
            for month_num in season:
                month_letter = month_dict[month_num]
                seas_str.append(month_letter)
                month_name = month_fullname_dict[month_num]
                seas_fullname_str.append(month_name)
            season_string = "".join(seas_str)
            season_fullname_string = "".join(seas_fullname_str)

            print(
                (
                    "Calculating {} for {}-{} {}".format(
                        metric, str(years[0]), str(years[1]), season_string
                    )
                )
            )

            # set up an iris constraint for the season
            season_constraint = iris.Constraint(
                time=lambda cell: cell.point.month in season
            )
            # year constraint
            year_constraint = iris.Constraint(
                time=lambda cell: years[0] <= cell.point.year <= years[1]
            )
            # extract the data matching the season and year constraints
            season_cube = cube.extract(year_constraint & season_constraint)

            # make sure season_cube exists
            if season_cube is None:
                raise Exception(
                    "Cube constriants of seas_mons and/or years do not match data in the input cube"
                )

            # calculate a time mean
            if metric == "mean":
                cube_stat = season_cube.collapsed("time", iris.analysis.MEAN)
            # calculate standard deviation (D.O.F=1)
            if metric == "std_dev":
                cube_stat = season_cube.collapsed("time", iris.analysis.STD_DEV)
            # calculate minimum
            if metric == "min":
                cube_stat = season_cube.collapsed("time", iris.analysis.MIN)
            # calculate maximum
            if metric == "max":
                cube_stat = season_cube.collapsed("time", iris.analysis.MAX)
            # calculate percentile
            if metric == "percentile":
                if not pc:
                    raise ValueError("percentile to calculate, pc, is not set.")
                if isinstance(pc, int):
                    cube_stat = season_cube.collapsed(
                        "time", iris.analysis.PERCENTILE, percent=pc
                    )
                else:
                    raise TypeError(
                        " pc must be an integer, it is currently {} of type {}".format(
                            pc, type(pc)
                        )
                    )

            # add a coord describing the season
            aux_seas = iris.coords.AuxCoord(
                season_string, long_name="season", units="no_unit"
            )
            cube_stat.add_aux_coord(aux_seas)

            aux_seas_fullname = iris.coords.AuxCoord(
                season_fullname_string, long_name="season_fullname", units="no_unit"
            )
            cube_stat.add_aux_coord(aux_seas_fullname)

            # add seasonal statistic of cube to cube list
            cube_list.append(cube_stat)

    return cube_list


def regular_point_to_rotated(cube, lon, lat):
    """
    Function to convert a regular lon lat point to
    the equivalent point on a rotated pole grid
    defined by the coordinate system of an input cube.
    Longitude can be either 0-360 or -180-180 degree.

    args
    ----
    cube: cube on rotated coord system
    lon: float of regular longitude point
    lat: float of regular latitude point

    Returns
    -------
    rot_lon: The equivalent longitude point on the grid of the input cube
    rot_lat: The equivalent latitude point on the grid of the input cube

    Notes
    -----


    An example:

    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> lat = 6.5
    >>> lon = 289 # on 0-360 degree
    >>> cube = iris.load_cube(file, 'air_temperature')
    >>> rot_lon, rot_lat = regular_point_to_rotated(cube, lon, lat)
    >>> print("{:.3f}".format(rot_lon), "{:.3f}".format(rot_lat))
    -84.330 3.336
    >>> lat = 6.5
    >>> lon = -71 # on -180-180 degree
    >>> rot_lon, rot_lat = regular_point_to_rotated(cube, lon, lat)
    >>> print("{:.3f}".format(rot_lon), "{:.3f}".format(rot_lat))
    -84.330 3.336
    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # get name of y coord
    ycoord = cube.coord(axis="Y", dim_coords=True)
    # cartopy.crs.RotatedGeodetic object
    rot_pole = cube.coord(ycoord).coord_system.as_cartopy_crs()
    # "regular" lon/lat coord system
    ll = ccrs.Geodetic()
    # Transform the lon lat point into rotated coordinates.
    rot_lon, rot_lat = rot_pole.transform_point(lon, lat, ll)

    return rot_lon, rot_lat


def rotated_point_to_regular(cube, rot_lon, rot_lat):
    """
    Function to convert a rotated lon lat point to
    the equivalent real lon lat point. Rotation of input point
    defined by the coordinate system of the input cube.
    Longitude output in -180-180 degree format.

    args
    ----
    cube: cube on rotated coord system, used as reference grid for transformation.
    rot_lon: float of rotated longitude point
    rot_lat: float of rotated latitude point

    Returns
    -------
    reg_lon: The equivalent real longitude point.
    reg_lat: The equivalent real latitude point.

    Notes
    -----

    An example:

    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube = iris.load_cube(file, 'air_temperature')
    >>> rot_lat = 3.34
    >>> rot_lon = -84.33
    >>> reg_lon, reg_lat = rotated_point_to_regular(cube, rot_lon, rot_lat)
    >>> print("{:.3f}".format(reg_lon), "{:.3f}".format(reg_lat))
    -71.003 6.502
    >>> rot_lat = 40
    >>> rot_lon = 370
    >>> reg_lon, reg_lat = rotated_point_to_regular(cube, rot_lon, rot_lat)
    >>> print("{:.3f}".format(reg_lon), "{:.3f}".format(reg_lat))
    116.741 82.265
    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # get name of y coord
    ycoord = cube.coord(axis="Y", dim_coords=True)
    # cartopy.crs.RotatedGeodetic object
    rot_pole = cube.coord(ycoord).coord_system.as_cartopy_crs()
    # "regular" lon/lat coord system
    ll = ccrs.Geodetic()
    # Transform the  rotated lon lat point into regular coordinates.
    reg_lon, reg_lat = ll.transform_point(rot_lon, rot_lat, rot_pole)

    return reg_lon, reg_lat


def windspeed(u_cube, v_cube):

    """
    This function calculates wind speed.

    args
    ----
    u_cube: cube of eastward wind
    v_cube: cube of northward wind

    Returns
    -------
    windspeed_cube: cube of wind speed in same units as input cubes.


    A simple example:

    >>> file = os.path.join(conf.DATA_DIR, 'gcm_monthly.pp')
    >>> u_cube = iris.load_cube(file, 'x_wind')
    >>> v_cube = iris.load_cube(file, 'y_wind')
    >>> ws = windspeed(u_cube, v_cube)
    >>> ws.attributes['formula']
    'sqrt(u**2, v**2)'
    >>> print(np.min(ws.data),np.max(ws.data))
    0.010195612 13.077045
    >>> ws.standard_name
    'wind_speed'
    """

    if not isinstance(u_cube, iris.cube.Cube) or not isinstance(v_cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    if u_cube.units != getattr(v_cube, "units", u_cube.units):
        raise ValueError(
            "units do not match, {} and {}".format(u_cube.units, v_cube.units)
        )

    # cube to put the windspeed in
    windspeed_cube = u_cube.copy()
    # adjust meta data
    windspeed_cube.standard_name = "wind_speed"
    if "STASH" in windspeed_cube.attributes:
        windspeed_cube.attributes.pop("STASH", None)
    windspeed_cube.attributes["formula"] = "sqrt(u**2, v**2)"

    windspeed_cube.data = np.sqrt(u_cube.data ** 2 + v_cube.data ** 2)

    return windspeed_cube


def wind_direction(u_cube, v_cube, unrotate=True):

    """
    Adapted from UKCP common_analysis.py
    (http://fcm1/projects/UKCPClimateServices/browser/trunk/code/)

    This function calculates the direction of the wind vector,
    which is the "to" direction (e.g. Northwards)
    (i.e. NOT Northerly, the "from" direction!)
    and the CF Convensions say it should be measured
    clockwise from North. If the data is on rotated pole, the
    rotate_winds iris function is applied as a default, but this
    can be turned off.

    Note: If you are unsure whether your winds need to be
    unrotated you can use http://www-nwp/umdoc/pages/stashtech.html
    and navigate to the relevant UM version and stash code:
        -    Rotate=0 means data is relative to the model grid and DOES need to be unrotated.
        -    Rotate=1 means data is relative to the lat-lon grid and DOES NOT need unrotating.

    args
    ----
    u_cube: cube of eastward wind
    v_cube: cube of northward wind
    unrotate: boolean, defaults to True. If true and data is rotated pole, the winds are unrotated, if set to False, they are not.

    Returns
    -------
    angle_cube: cube of wind direction in degrees (wind direction 'to' not 'from')


    A simple example:

    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> u_cube = iris.load_cube(file, 'x_wind')
    >>> v_cube = iris.load_cube(file, 'y_wind')
    >>> angle = wind_direction(u_cube, v_cube)
    data is on rotated coord system, un-rotating . . .
    >>> angle.attributes['formula']
    '(-(arctan2(v, u)*180/pi)+90)%360'
    >>> print(np.min(angle.data),np.max(angle.data))
    0.0181427 359.99008
    >>> angle.standard_name
    'wind_to_direction'
    >>> # Now take the same data and calculate the wind direction WITHOUT rotating
    ... angle_unrot = wind_direction(u_cube, v_cube, unrotate=False)
    >>> # Print the difference between wind direction data
    ... # where unrotate=True unrotate=False to illustrate
    ... # the importance of getting the correct unrotate option
    ... print(angle.data[0,100,:10] - angle_unrot.data[0,100,:10])
    [-26.285324 -26.203125 -26.120728 -26.038239 -25.955551 -25.872742
     -25.78981  -25.70668  -25.623428 -25.539993]
    """

    if u_cube.units != getattr(v_cube, "units", u_cube.units):
        raise ValueError(
            "units do not match, {} and {}".format(u_cube.units, v_cube.units)
        )

    # check if data is on rotated pole, unrotate if necessary
    cs_str = str(u_cube.coord_system())
    if cs_str.find("Rotated") != -1:
        if unrotate:
            print("data is on rotated coord system, un-rotating . . .")
            target_cs = iris.coord_systems.GeogCS(iris.fileformats.pp.EARTH_RADIUS)
            u_cube, v_cube = iris.analysis.cartography.rotate_winds(
                u_cube, v_cube, target_cs
            )

    # create a cube for the output and adjust meta data
    angle_cube = u_cube.copy()
    angle_cube.units = "degree"
    angle_cube.standard_name = "wind_to_direction"
    angle_cube.long_name = "wind vector direction"
    angle_cube.var_name = "angle"
    if "STASH" in angle_cube.attributes:
        angle_cube.attributes.pop("STASH", None)
    angle_cube.attributes[
        "direction"
    ] = "Angle of wind vector measured clockwise from Northwards"
    angle_cube.attributes["formula"] = "(-(arctan2(v, u)*180/pi)+90)%360"

    angle_cube.data = np.arctan2(v_cube.data, u_cube.data) * 180.0 / np.pi
    # That gives the angle of the vector in degrees, anticlockwise from Eastwards.
    # (in the range -180 to +180)
    # But we want the bearing clockwise from Northwards (0,360), so:
    angle_cube.data = (-angle_cube.data + 90.0) % 360

    return angle_cube


if __name__ == "__main__":
    doctest.testmod()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# (C) British Crown Copyright 2017-2020 Met Office.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# -----------------------------------------------------------------------------

import iris
import iris.analysis
import numpy as np
from six import string_types, integer_types
import iris.coord_categorisation as iccat
import doctest
import os.path

import catnip.config as conf
import iris.exceptions
from dask import array as da

def _get_xy_noborder(mask):
    '''
    make  a function that returns the indices
    of where the mask is valid. If the mask is all True (all masked)
    raises a ValueError

    args
    ----
    mask: mask from numpy array

    Returns
    -------
    x1, x2, y1, y2: int giving space where the data is valid

    '''

    if np.all(mask):
        raise ValueError("All values masked - can't get indices")
    ys, xs = np.where(~mask)
    x1 = min(xs)
    x2 = max(xs) + 1
    y1 = min(ys)
    y2 = max(ys) + 1

    return x1, x2, y1, y2


def add_aux_unrotated_coords(cube):
    """
    This function takes a cube that is on a rotated pole
    coordinate system and adds to it, two addtional
    auxillary coordinates to hold the unrotated coordinate
    values.

    args
    ----
    cube: iris cube on an rotated pole coordinate system

    Returns
    -------
    cube: input cube with auxilliary coordinates of unrotated
    latitude and longitude

    Notes
    -----

    See below for an example that should be run with python3:

    >>> file = os.path.join(conf.DATA_DIR, 'mslp.daily.rcm.viet.nc')
    >>> cube = iris.load_cube(file)
    >>> print([coord.name() for coord in cube.coords()])
    ['time', 'grid_latitude', 'grid_longitude']
    >>> auxcube = add_aux_unrotated_coords(cube)
    >>> print([coord.name() for coord in auxcube.coords()])
    ['time', 'grid_latitude', 'grid_longitude', 'latitude', 'longitude']
    >>> print(auxcube.coord('latitude'))
    AuxCoord(array([[35.32243855, 35.33914928, 35.355619  , ..., 35.71848081,
            35.70883111, 35.69893388],
           [35.10317609, 35.11986604, 35.13631525, ..., 35.49871728,
            35.48908   , 35.47919551],
           [34.88390966, 34.90057895, 34.91700776, ..., 35.27895246,
            35.26932754, 35.25945571],
           ...,
           [ 6.13961446,  6.15413611,  6.16844578, ...,  6.48307389,
             6.47472284,  6.46615667],
           [ 5.92011032,  5.93461779,  5.94891347, ...,  6.26323044,
             6.25488773,  6.24633011],
           [ 5.70060768,  5.71510098,  5.72938268, ...,  6.04338876,
             6.03505439,  6.02650532]]), standard_name=None, units=Unit('degrees'), long_name='latitude')
    >>> print(auxcube.shape)
    (360, 136, 109)
    >>> print(auxcube.coord('latitude').shape)
    (136, 109)
    >>> print(auxcube.coord('longitude').shape)
    (136, 109)

    """

    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # get cube's coordinate system
    cs = cube.coord_system()

    if str(cs).find("Rotated") == -1:
        raise TypeError(
            "The cube is not on a rotated pole, coord system is {}".format(str(cs))
        )

    auxcube = cube.copy()
    # get coord names
    # Longitude
    xcoord = auxcube.coord(axis="X", dim_coords=True)
    # Latitude
    ycoord = auxcube.coord(axis="Y", dim_coords=True)

    # read in the grid lat/lon points from the cube
    glat = auxcube.coord(ycoord).points
    glon = auxcube.coord(xcoord).points

    # create a rectangular grid out of an array of
    # glon and glat values, shape will be len(glat)xlen(glon)
    x, y = np.meshgrid(glon, glat)

    # get the cube dimensions which corresponds to glon and glat
    x_dim = auxcube.coord_dims(xcoord)[0]
    y_dim = auxcube.coord_dims(ycoord)[0]

    # define two new variables to hold the unrotated coordinates
    rlongitude, rlatitude = iris.analysis.cartography.unrotate_pole(
        x, y, cs.grid_north_pole_longitude, cs.grid_north_pole_latitude
    )

    # create two new auxillary coordinates to hold
    # the values of the unrotated coordinates
    reg_long = iris.coords.AuxCoord(rlongitude, long_name="longitude", units="degrees")
    reg_lat = iris.coords.AuxCoord(rlatitude, long_name="latitude", units="degrees")

    # add two auxilary coordinates to the cube holding
    # regular(unrotated) lat/lon values
    auxcube.add_aux_coord(reg_long, [y_dim, x_dim])
    auxcube.add_aux_coord(reg_lat, [y_dim, x_dim])

    return auxcube


def add_bounds(cube, coord_names, bound_position=0.5):
    """
        Simple function to check whether a
        coordinate in a cube has bounds, and
        add them if it doesn't.

        args
        ----
        cube: iris cube
        coord_names: string or list of strings containing the name/s
                     of the coordinates you want to add bounds to.
        bound_position: Optional, the desired position of the bounds relative to
                        the position of the points. Default is 0.5.

        Returns
        -------
        cube: cube with bounds added

        Notes
        -----
        Need to be careful that it is appropriate
        to add bounds to the data, e.g. if data
        are instantaneous, time bounds are not
        appropriate.

        An example:


        >>> file = os.path.join(conf.DATA_DIR, 'mslp.daily.rcm.viet.nc')
        >>> cube = iris.load_cube(file)
        >>> bcube = add_bounds(cube, 'time')
        time coordinate already has bounds, none will be added
        >>> bcube = add_bounds(cube, 'grid_latitude')
        grid_latitude bounds added
        >>> bcube = add_bounds(cube, ['grid_latitude','grid_longitude'])
        grid_latitude bounds added
        grid_longitude bounds added
        """

    # check if the input is an Iris cube
    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # check if the coordinate name input is a string
    if not isinstance(coord_names, (string_types, list)):
        raise TypeError("Input coordinate must be a string")

    bcube = cube.copy()

    # find names of dim coords
    c_names = [c.name() for c in bcube.coords()]

    # if coord_names is a single string, it will be split,
    # by the loop this statement checks for that case and
    # puts stash into a tuple to prevent splitting.
    if isinstance(coord_names, string_types):
        coord_names = tuple([coord_names])

    for coord in coord_names:

        # check if coord is a string
        if not isinstance(coord, string_types):
            raise TypeError(
                "Coordinate {} must be a string, it is currently a {}".format(
                    str(coord), type(coord)
                )
            )

        # check coord is a coordinate of the cube
        if coord not in c_names:
            raise AttributeError(
                "{} is not a coordinate, available coordinates are: {}".format(
                    coord, c_names
                )
            )

        # check if the coord already has bounds
        if bcube.coord(coord).has_bounds():
            print(
                ("{} coordinate already has bounds, none will be added".format(coord))
            )

        # add bounds to coord
        else:
            bcube.coord(coord).guess_bounds(bound_position=bound_position)
            print(("{} bounds added".format(coord)))

    return bcube


def add_coord_system(cube):
    """
    A cube must have a coordinate system in order to be regridded.

    This function checks whether a cube has a coordinate system. If
    the cube has no coordinate system, the standard the ellipsoid
    representation wgs84 (ie. the one used by GPS) is added.

    Note: It will not work for rotated pole data without a
    coordinate system.

    args
    ----
    cube: iris cube

    Returns
    -------
    cube: The copy of the input cube with coordinate system added, if the cube didn't have one already.

    Notes
    -----

    A simple example:

    >>> file = os.path.join(conf.DATA_DIR, 'gtopo30_025deg.nc')
    >>> cube = iris.load_cube(file)
    >>> print(cube.coord('latitude').coord_system)
    None
    >>> cscube = add_coord_system(cube)
    Coordinate system  GeogCS(6371229.0) added to cube
    >>> print(cscube.coord('latitude').coord_system)
    GeogCS(6371229.0)
    """

    # Note: wgs84 is the World Geodetic System, and a standard coord
    # system in iris. In GeogCS(6371229.0), 6371229 is the Earth's
    # radius in m. See:
    # https://scitools.org.uk/iris/docs/v1.9.0/html/iris/iris/coord_systems.html

    # check if the input is an Iris cube
    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    cscube = cube.copy()
    cs = cscube.coord_system()

    if cs is not None:
        if str(cs).find("Rotated") == 0:
            # not possible to add a coord system for
            # rotated pole cube without knowing the
            # rotation. Give error message.
            raise TypeError("Error, no coordinate system for rotated pole cube")
    else:
        coord_names = [coord.name() for coord in cscube.coords(dim_coords=True)]
        wgs84_cs = iris.coord_systems.GeogCS(6371229.0)
        if "latitude" in coord_names:
            cscube.coord("latitude").coord_system = wgs84_cs
        if "longitude" in coord_names:
            cscube.coord("longitude").coord_system = wgs84_cs
        print("Coordinate system  GeogCS(6371229.0) added to cube")

    return cscube


def add_time_coord_cats(cube):
    """
    This function takes in an iris cube, and adds a range of
    numeric co-ordinate categorisations to it. Depending
    on the data, not all of the coords added will be relevant.

    args
    ----
    cube: iris cube that has a coordinate called 'time'

    Returns
    -------
    Cube: cube that has new time categorisation coords added

    Notes
    -----
    test

    A simple example:

    >>> file = os.path.join(conf.DATA_DIR, 'mslp.daily.rcm.viet.nc')
    >>> cube = iris.load_cube(file)
    >>> coord_names = [coord.name() for coord in cube.coords()]
    >>> print((', '.join(coord_names)))
    time, grid_latitude, grid_longitude
    >>> ccube = add_time_coord_cats(cube)
    >>> coord_names = [coord.name() for coord in ccube.coords()]
    >>> print((', '.join(coord_names)))
    time, grid_latitude, grid_longitude, day_of_month, day_of_year, month, month_number, season, season_number, year
    >>> # print every 50th value of the added time cat coords
    ... for c in coord_names[3:]:
    ...     print(ccube.coord(c).long_name)
    ...     print(ccube.coord(c).points[::50])
    ...
    day_of_month
    [ 1 21 11  1 21 11  1 21]
    day_of_year
    [  1  51 101 151 201 251 301 351]
    month
    ['Jan' 'Feb' 'Apr' 'Jun' 'Jul' 'Sep' 'Nov' 'Dec']
    month_number
    [ 1  2  4  6  7  9 11 12]
    season
    ['djf' 'djf' 'mam' 'jja' 'jja' 'son' 'son' 'djf']
    season_number
    [0 0 1 2 2 3 3 0]
    year
    [2000 2000 2000 2000 2000 2000 2000 2000]

    """

    # most errors pop up when you try to add a coord that has
    # previously been added, or the cube doesn't contain the
    # necessary attribute.

    ccube = cube.copy()

    # numeric
    try:
        iccat.add_day_of_year(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    try:
        iccat.add_day_of_month(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    try:
        iccat.add_month_number(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    try:
        iccat.add_season_number(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    try:
        iccat.add_year(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    # strings
    try:
        iccat.add_month(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    try:
        iccat.add_season(ccube, "time")
    except AttributeError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))
    except ValueError as err:
        print(("add_time_coord_cats: {}, skipping . . . ".format(err)))

    return ccube

def extract_rot_cube(cube, min_lat, min_lon, max_lat, max_lon):
    """
    Function etracts the specific region from the cube.
    args
    ----
    cube: cube on rotated coord system, used as reference grid for transformation.
    Returns
    -------
    min_lat: The minimum latitude point of the desired extracted cube.
    min_lon: The minimum longitude point of the desired extracted cube.
    max_lat: The maximum latitude point of the desired extracted cube.
    max_lon: The maximum longitude point of the desired extracted cube.
    An example:
    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube = iris.load_cube(file, 'air_temperature')
    >>> min_lat = 50
    >>> min_lon = -10
    >>> max_lat = 60
    >>> max_lon = 0
    >>> extracted_cube = extract_rot_cube(cube, min_lat, min_lon, max_lat, max_lon)
    >>> print(np.max(extracted_cube.coord('latitude').points))
    61.47165097005264
    >>> print(np.min(extracted_cube.coord('latitude').points))
    48.213032844268646
    >>> print(np.max(extracted_cube.coord('longitude').points))
    3.642576550089792
    >>> print(np.min(extracted_cube.coord('longitude').points))
    -16.385571344717235
    """

    # adding unrotated coords to the cube
    cube = add_aux_unrotated_coords(cube)

    # mask the cube using the true lat and lon
    lats = cube.coord('latitude').points
    lons = cube.coord('longitude').points
    select_lons = (lons >= min_lon) & (lons <= max_lon)
    select_lats = (lats >= min_lat) & (lats <= max_lat)
    selection = select_lats & select_lons
    selection = da.broadcast_to(selection, cube.shape)
    cube.data = da.ma.masked_where(~selection, cube.core_data())


    # grab a single 2D slice of X and Y and take the mask
    lon_coord = cube.coord(axis='X', dim_coords=True)
    lat_coord = cube.coord(axis='Y', dim_coords=True)
    for yx_slice in cube.slices(['grid_latitude', 'grid_longitude']):
        cmask = yx_slice.data.mask
        break

    # now cut the cube down along X and Y coords
    x1, x2, y1, y2 = _get_xy_noborder(cmask)
    idx = len(cube.shape) * [slice(None)]

    idx[cube.coord_dims(cube.coord(axis='x', dim_coords=True))[0]] = slice(x1, x2, 1)
    idx[cube.coord_dims(cube.coord(axis='y', dim_coords=True))[0]] = slice(y1, y2,1)

    extracted_cube = cube[tuple(idx)]

    return extracted_cube

def remove_forecast_coordinates(iris_cube):
    """A function to remove the forecast_period and forecast_reference_time coordinates from the UM PP files

    args
    ----
    iris_cube: input iris_cube

    Returns
    -------
    iris_cube: iris cube without the forecast_period and forecast_reference_time coordinates

    Notes
    -----

    See below for examples:

    >>> cube_list_fcr = iris.cube.CubeList()
    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube_list = iris.load(file)
    >>> for cube in cube_list:
    ...     cube_fcr = remove_forecast_coordinates(cube)
    ...     cube_list_fcr.append(cube_fcr)
    Removed the forecast_period coordinate from Heavyside function on pressure levels cube
    Removed the forecast_reference_time coordinate from Heavyside function on pressure levels cube
    Removed the forecast_period coordinate from air_temperature cube
    Removed the forecast_reference_time coordinate from air_temperature cube
    Removed the forecast_period coordinate from relative_humidity cube
    Removed the forecast_reference_time coordinate from relative_humidity cube
    Removed the forecast_period coordinate from specific_humidity cube
    Removed the forecast_reference_time coordinate from specific_humidity cube
    Removed the forecast_period coordinate from x_wind cube
    Removed the forecast_reference_time coordinate from x_wind cube
    Removed the forecast_period coordinate from y_wind cube
    Removed the forecast_reference_time coordinate from y_wind cube

    Now check if the forecast coordinates have been removed

    >>> for cube in cube_list_fcr:
    ...     cube_nfc = remove_forecast_coordinates(cube)
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
    'Expected to find exactly 1 forecast_period coordinate, but found none.'
    'Expected to find exactly 1 forecast_reference_time coordinate, but found none.'
"""

    try:
        iris_cube.remove_coord("forecast_period")
        print(
            (
                "Removed the forecast_period coordinate from {} cube".format(
                    iris_cube.name()
                )
            )
        )
    except iris.exceptions.CoordinateNotFoundError as coord_not_found:
        print("{}".format(coord_not_found))
    try:
        iris_cube.remove_coord("forecast_reference_time")
        print(
            (
                "Removed the forecast_reference_time coordinate from {} cube".format(
                    iris_cube.name()
                )
            )
        )
    except iris.exceptions.CoordinateNotFoundError as coord_not_found:
        print("{}".format(coord_not_found))

    return iris_cube


def rim_remove(cube, rim_width):
    """ Return IRIS cube with rim removed.

    args
    ----
    cube: input iris cube
    rim_width: integer, number of grid points to remove from edge of lat and long

    Returns
    -------
    rrcube: rim removed cube

    Notes
    -----

    See below for examples:

    >>> cube_list_rr = iris.cube.CubeList()
    >>> file = os.path.join(conf.DATA_DIR, 'rcm_monthly.pp')
    >>> cube_list = iris.load(file)
    >>> for cube in cube_list:
    ...     cube_rr = rim_remove(cube, 8)
    ...     cube_list_rr.append(cube_rr)
    ...
    Removed 8 size rim from Heavyside function on pressure levels
    Removed 8 size rim from air_temperature
    Removed 8 size rim from relative_humidity
    Removed 8 size rim from specific_humidity
    Removed 8 size rim from x_wind
    Removed 8 size rim from y_wind
    >>> file = os.path.join(conf.DATA_DIR, 'rcm_mslp_monthly.pp')
    >>> mslp_cube = iris.load_cube(file)
    >>>
    >>> mslp_cube_rr = rim_remove(mslp_cube, 8)
    Removed 8 size rim from air_pressure_at_sea_level
    >>>
    >>> print(len(mslp_cube.coord('grid_latitude').points))
    432
    >>> print(len(mslp_cube.coord('grid_longitude').points))
    444
    >>> print(len(mslp_cube.coord('grid_latitude').points))
    432
    >>> print(len(mslp_cube.coord('grid_longitude').points))
    444
    >>>
    >>> mslp_cube_rrrr = rim_remove(mslp_cube_rr, 8)
    WARNING - This cube has already had it's rim removed
    Removed 8 size rim from air_pressure_at_sea_level


    Now test for failures:

    >>> mslp_cube_rr = rim_remove(cube, 8.2) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: Please provide a positive integer for rim_width
    >>> mslp_cube_rr = rim_remove(cube, -5) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    IndexError: Please provide a positive integer > 0 for rim_width
    >>> mslp_cube_rr = rim_remove(cube, 400) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    IndexError: length of lat or lon coord is < rim_width*2
    >>> mslp_cube_rr = rim_remove(cube, 0) # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    IndexError: Please provide a positive integer > 0 for rim_width
    >>> mslp_cube_rr = rim_remove(cube, 'a') # doctest: +ELLIPSIS
    Traceback (most recent call last):
    ...
    TypeError: Please provide a positive integer for rim_width
    """
    # check if the input is an Iris cube
    if not isinstance(cube, iris.cube.Cube):
        raise TypeError("Input is not a cube")

    # check whether rim_width is an integer
    if not isinstance(rim_width, (integer_types)):
        raise TypeError("Please provide a positive integer for rim_width")
    if rim_width <= 0:
        raise IndexError("Please provide a positive integer > 0 for rim_width")

    # check whether this cube has already had it's rim removed
    if "rim_removed" in cube.attributes:
        print("WARNING - This cube has already had it's rim removed")

    # Longitude
    xcoord = cube.coord(axis="X", dim_coords=True)
    # Latitude
    ycoord = cube.coord(axis="Y", dim_coords=True)

    # make sure specified rim_width is going to work
    if len(xcoord.points) <= (rim_width * 2) or len(ycoord.points) <= (rim_width * 2):
        raise IndexError("length of lat or lon coord is < rim_width*2")

    # Remove rim from Longitude
    rrcube = cube.subset(xcoord[rim_width : -1 * rim_width])
    # Remove rim from Latitude
    rrcube = rrcube.subset(ycoord[rim_width : -1 * rim_width])
    # add meta data that rim has been removed
    rrcube.attributes["rim_removed"] = "{} point rim removed".format(rim_width)

    print(("Removed {} size rim from {}".format(rim_width, cube.name())))

    return rrcube


if __name__ == "__main__":
    doctest.testmod()



import os.path
import configparser


# The full path to the "catnip" package.
ROOT_PATH = os.path.abspath(os.path.dirname(__file__))

# The full path of test data directory
DATA_DIR = os.path.join(ROOT_PATH, "data")

# The full path of the KGO dir
KGO_DIR = os.path.join(ROOT_PATH, "kgo")



import setuptools
from distutils.core import setup



with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='mo-catnip',
      version='0.0.5',
      description='Climate analysis tool',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Met Office',
      author_email='catnip@metoffice.gov.uk',
      license = "BSD",
      url='https://github.com/MetOffice/CATNIP',
      packages=setuptools.find_packages(where='lib'),
      package_dir={"": "lib"},
      python_requires='>=3.6',
      install_requires=['numpy',
                        'matplotlib',
                        'six',
                        'scipy'
                        ],
      keywords = ["cmip", "climate", "analysis", "rcp"],
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Development Status :: 3 - Alpha",
     ]
)
